package com.loafle.overflow.crawler.jmx;

import com.loafle.overflow.crawler.Crawler;
import com.loafle.overflow.crawler.config.*;
import com.loafle.overflow.crawler.result.OFResultSet;

import javax.management.*;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;
import javax.management.remote.rmi.RMIConnectorServer;

import javax.naming.Context;
import javax.rmi.ssl.SslRMIClientSocketFactory;
import java.io.IOException;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;

/**
 * Created by geek on 2017-04-06.
 */

public class JmxCrawler extends Crawler{
    private static final Logger logger = Logger.getLogger(JmxCrawler.class.getName());

    private String jmxUrl;
    private String username;
    private String password;
    private boolean ssl;

    public JmxCrawler() {}

    private List<String> parse(Map<String, String> dataMap, Map<String, Integer> meta) {

        List<String> row =  new ArrayList<>(Arrays.asList(new String[meta.size()]));
        for (Map.Entry<String, Integer> info : meta.entrySet()) {
            String data = dataMap.get(info.getKey());
            data = data.trim();
            row.set(info.getValue().intValue(),data);
        }
        return row;
    }

    private Map<String, String> parseAttribute(MBeanServerConnection beanCon, ObjectName mbeanName) throws IntrospectionException, ReflectionException, InstanceNotFoundException, IOException, AttributeNotFoundException, MBeanException {

        Map<String, String> returnMap = new HashMap<>();
        MBeanInfo info = beanCon.getMBeanInfo(mbeanName);

        for (MBeanAttributeInfo attr : info.getAttributes()) {

            if (!attr.isReadable()) {
                logCrawler(mbeanName, attr, "not readable");
                continue;
            }
            Object att = beanCon.getAttribute(mbeanName, attr.getName());
            if (att != null)
                returnMap.put(attr.getName(), att.toString());
        }
        return returnMap;
    }

    @Override
    public Object getInternal(Config config) throws Exception {

        String hostIp = config.getTarget().getConnection().getIp();
        String port = config.getTarget().getConnection().getPort();

        this.jmxUrl = "service:jmx:rmi:///jndi/rmi://" + hostIp + ":" +port + "/jmxrmi";
        this.username = null;
        this.password = null;

        // connection
        MBeanServerConnection beanCon;
        JMXConnector jmxconn = null;
        try {


            Map<String, Object> environment = getAuthority();
            jmxconn = JMXConnectorFactory.connect(new JMXServiceURL(this.jmxUrl), environment);
            beanCon = jmxconn.getMBeanServerConnection();

            List<OFResultSet> resultSetList = new ArrayList<>();
            List<ObjectName> objectNames = new LinkedList<ObjectName>();
            objectNames.add(null);
            ObjectName on = null;

            for (Item item : config.getItems()) {

                OFResultSet resultSet = OFResultSet.newInstance(item);

                // get object
                String query = item.getQueryInfo().getQuery();
                on = new ObjectName(query);
                Set<ObjectInstance> names = beanCon.queryMBeans(on, null);

                List<Map<String,Object>> keys = (List<Map<String,Object>>) item.getQueryInfo().getExtend().get("aliases");

                for (ObjectInstance name : names) {

                    Map<String, String> attrs = parseAttribute(beanCon,name.getObjectName());
                    if (keys != null) {
                        for ( Map<String,Object> keyMap : keys ) {
                            String key = (String) keyMap.get("key");
                            int index =  (int) keyMap.get("index");
                            String objectName = parseObjectName(name.toString(),key);
                            String alias = item.getMappingInfo().getArrayColumns().get(index);
                            attrs.put(alias,objectName);
                        }
                    }

                    resultSet.addRow(parse(attrs, resultSet.getMeta()));
                }

                resultSetList.add(resultSet);
            }

            return resultSetList;
        } catch(Exception e) {
            e.printStackTrace();
            throw e;
        } finally {
            if (jmxconn != null) {
                jmxconn.close();
            }
        }
    }

    private String parseObjectName(String s, String key) {

        int start = s.indexOf("[");
        int end = s.lastIndexOf("]");
        s = s.substring(start+1,end);

        String [] slices = s.split(",");
        String returnData = "";
        for (String item : slices) {
            String [] itemSplit = item.split("=");
            if (itemSplit == null || itemSplit.length != 2) {continue;}
            if (itemSplit[0].equals(key)) {
                returnData = itemSplit[1];
            }
        }

        return returnData;
    }

    private Map<String, Object> getAuthority() {

        Map<String, Object> environment = new HashMap<String, Object>();

        if (this.username != null && this.username.length() != 0 && this.password != null && this.password.length() != 0) {
            String[] credent = new String[] {this.username, this.password};
            environment.put(JMXConnector.CREDENTIALS, credent);
        }

        if (this.ssl) {
            environment.put(Context.SECURITY_PROTOCOL, "ssl");
            SslRMIClientSocketFactory clientSocketFactory = new SslRMIClientSocketFactory();
            environment.put(RMIConnectorServer.RMI_CLIENT_SOCKET_FACTORY_ATTRIBUTE, clientSocketFactory);
            environment.put("com.sun.jndi.rmi.factory.socket", clientSocketFactory);
        }

        return environment;
    }

    private static void logCrawler(String name, String msg) {
        logger.log(Level.FINE, "crawler : '" + name + "': " + msg);
    }

    private static void logCrawler(ObjectName mbeanName, MBeanAttributeInfo attr, String msg) {
        logCrawler(mbeanName + "'_'" + attr.getName(), msg);
    }

}
